/* LineGroup.java
 *
 * created: Mon Oct 12 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/uk/ac/sanger/pathogens/embl/LineGroup.java,v 1.2 1999/06/15 10:12:03 kmr Exp $
 */

package uk.ac.sanger.pathogens.embl;

import java.io.Reader;
import java.io.Writer;
import java.io.BufferedReader;
import java.io.IOException;

/**
 *  This class corresponds to a group of associated lines in an EMBL entry.
 *  An example of a group of associated lines is all the lines in an entry
 *  that start with FT.
 *
 *  @author Kim Rutherford
 *  @version $Id: LineGroup.java,v 1.2 1999/06/15 10:12:03 kmr Exp $
 *
 */

abstract class LineGroup
    extends EMBLObject {
  /**
   *  Try to read and return a new LineGroup object from a stream.
   *  @param entry The Entry that owns this object.
   *  @param reader The stream to read from.
   *  @return A new LineGroup object or null if stream is at the end of file.
   *  @exception IOException Thrown if exception occurs while reading.
   *  @exception ReadFormatException Thrown if the format of the input is in
   *    error.
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public static LineGroup readNextLineGroup (DocumentEntry entry,
                                             LinePushBackReader reader)
      throws IOException, InvalidRelationException {

    String line;
    
    line = reader.readLine ();

    if (line == null) {
      // end of file
      return null;
    } else {

//       if (line.length () >= 5 && ! line.substring (2).startsWith ("   ")) {
//         // a line of a embl file can only have space characters in columns 3,
//         // 4 and 5 of each line.
//         throw new ReadFormatException ("Failed to read embl data " +
//                                        "- columns 3, 4 and 5 must be empty",
//                                        reader.getLineNumber ());
//       }

      final int line_type = LineGroup.getLineType (line);

      reader.pushBack (line);
      
      switch (line_type) {
      case SEQUENCE:
//        System.out.println ("in getNextLineGroup (): reading SQ");
        return StreamSequenceFactory.makeStreamSequence (reader);
        
      case FEATURE_TABLE:
//        System.out.println ("in getNextLineGroup (): reading FT");
        return new FeatureTable (entry, reader);

      case END_OF_ENTRY:
        // in this case we do want to read the line (which will be //) so that
        // the next call to readNextEntry () starts on the next entry
        reader.readLine ();
        return null;
        
      default:
        return new Unknown (reader);
      }
    }
  }

  /**
   *  Return the embl line type of the line contained in the argument String.
   */
  public static int getLineType (String line) {
    if (line.startsWith (FEATURE_TABLE_STRING)) {
      return FEATURE_TABLE;
    }
    if (line.startsWith (END_OF_ENTRY_STRING)) {
      return END_OF_ENTRY;
    }
    if (line.startsWith ("  ") || line.startsWith (SEQUENCE_STRING)) {
     return SEQUENCE;
    }
    if (line.length () >= 2 &&
        Character.isLetter (line.charAt (0)) &&
        Character.isLetter (line.charAt (1)) &&
        (line.length () == 2 ||
         line.length () == 3 && line.endsWith (" ") ||
         line.length () == 4 && line.endsWith ("  ") ||
         line.length () >= 5 && line.substring (2,5).equals ("   "))) {
      // this covers all the lines in the header
      return UNKNOWN;
    }
    // default is raw sequence
    return SEQUENCE;
  }


  /**
   *  Returns a String containing the contents of the line with the initial
   *  type string (two letters) and white space (three spaces) removed.
   */
  public static String getRestOfLine (String line) {
    final int END_OF_SPACES = 5;

//    System.out.println ("in getRestOfLine (): " + line);
    if (line.length () > END_OF_SPACES) {
//      System.out.println ("in getRestOfLine () returning: " +
//                          line.substring (END_OF_SPACES));
      return line.substring (END_OF_SPACES);
    } else {
      return "";
    }
  }

  /**
   *  Write the end of entry marker - "//".
   **/
  public static void writeEndOfEMBLEntry (Writer writer) throws IOException {
    writer.write (END_OF_ENTRY_STRING + "\n");
  }
  
  /**
   *  Return the type of this LineGroup object (SEQUENCE, FEATURE_TABLE, etc.)
   **/
  public abstract int getType ();

  /**
   *  Write this object to the given stream.  The output will be in EMBL
   *  format (but see the diana_extensions argument).
   *  @param writer The stream to write to.
   *  @param diana_extensions If true then the any diana additions to the embl
   *    file format will be included in the output, otherwise they will be
   *    removed.
   **/
  public abstract void writeToStream (final Writer out_stream,
                                      final boolean diana_extensions)
      throws IOException;
  
  /**
   *  
   **/
  final static int UNKNOWN = 0;

  /**
   *  //
   **/
  final static int END_OF_ENTRY = 1;
  final static String END_OF_ENTRY_STRING = "//";

  /**
   *  SQ
   **/
  final static int SEQUENCE = 2;
  final static String SEQUENCE_STRING = "SQ";

  /**
   *  FT
   **/
  final static int FEATURE_TABLE = 3;
  final static String FEATURE_TABLE_STRING = "FT   ";
}


